#ifndef __THREAD_H__
#define __THREAD_H__

// $Id: TThread.h,v 1.14 1999/03/02 19:16:52 stanis Exp $

#include <queue>
#include <map>

#if defined(WIN32)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
using namespace std;
#undef DrawText
#else
#define PTHREAD
#define _REENTRANT
#include <unistd.h>
#include <pthread.h>
#endif


class ThreadException
{
  public: 
	enum THREAD_ERROR_CODE { COND_CREATE, COND_WAIT, COND_SIGNAL, 
							 COND_BCAST, COND_DESTROY, MUTEX_INIT, 
							 MUTEX_LOCK, MUTEX_UNLOCK, MUTEX_DESTROY, 
							 MUTEX_TRYLOCK, THREAD_CREATE, THREAD_JOIN, 
							 THREAD_DISOWN, THREAD_KILL, THREAD_ATTRINIT, 
							 THREAD_SETSCOPE  };

	ThreadException(char *why, THREAD_ERROR_CODE code);

	virtual ~ThreadException();
	void Handle();

	THREAD_ERROR_CODE GetCode() const { return m_code; }
  private:
	char *m_why;
	THREAD_ERROR_CODE m_code;
};

class Message
{
  public:
	Message() {}
	virtual ~Message() {}

  private:
	
};

class Conditional;

class Mutex
{
	friend class Conditional;
  public:
	Mutex();
	virtual ~Mutex();
	
	void Lock();
	void Unlock();
	bool TryLock();
  private:
  	Mutex(const Mutex &m);			// Not Implemented
	Mutex &operator=(const Mutex &m);	// Not Implemented
#ifdef PTHREAD
	pthread_mutex_t m_mutex;
#elif defined(WIN32)
	LONG m_mutex;
#endif
};

class Conditional
{
  public:
	Conditional();
	virtual ~Conditional();
	void Wait(Mutex *mutex);
	void Signal();
	void Broadcast();
  private:
  	Conditional(const Conditional &c);		// Not Implemented
	Conditional &operator=(const Conditional &c); 	// Not Implemented
#ifdef PTHREAD
	pthread_cond_t m_cond;
#elif defined(WIN32)
	queue<Mutex *> m_wait_list;
	Mutex m_wait_list_mutex;
#endif
};

class Messenger
{
  public:
	Messenger();
	virtual ~Messenger();
	void SndMessage(Message *msg);
	Message *GetMessage(bool block);

  private:
	Conditional m_message_queue_cond;
	Mutex m_message_queue_mutex;
	queue<Message *> m_message_queue;
};


class Thread : public Messenger
{
  public:
	// Initalize a new thread structure
	Thread();

	// Reap the thread if it hasn't been disowned, otherwise outright kill
	// it.
	virtual ~Thread();

	// Start the thread running
	void Run();

	// Wait for the thread to exit, if not disowned
	void Reap();

	// Disown the thread so that we won't ever wait until it is finished
	void Disown();

	// Kill the thread dead.
	void Kill();

	// Suspend the thread for a while
	static void Sleep( int sec, int usec );

	// Debugging functions.  These are UNSUPORTED
#ifdef PTHREAD
	static pthread_t GetThreadID() { return pthread_self(); }
#elif defined(WIN32)
	static DWORD GetThreadID() { return GetCurrentThreadId(); }
#endif

	bool Running();
  protected:
	// Overide this with your thread main loop.
	virtual void *ThreadMain();	

  private:
  	Thread(const Thread &t); 		// Not Implemented
	Thread &operator=(const Thread &t);	// Not Implemented

	bool m_disowned;
	bool m_reaped;
	bool m_killed;
	bool m_running;

	static void *thread_func(void *);
	static void cleanup_handler(void *);


#ifdef PTHREAD
	static map<Thread *, Mutex *> ms_thread_lock;
	pthread_t m_child_thread_id;
	pthread_t m_parent_thread_id;
#elif defined(WIN32)
	DWORD   m_child_thread_id;
	DWORD   m_parent_thread_id;

	HANDLE  m_child_thread_handle;
	HANDLE  m_parent_thread_handle;
#endif
};

#endif


