// Implementation of TThread classes on all platforms
// $Id: Thread.cpp,v 1.17 1999/03/02 19:16:52 stanis Exp $

#include "TThread.h"

#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <algorithm>

#ifdef PTHREAD
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#endif

#ifdef _DEBUG
#define ASSERT(x) assert(x)
#else
#define ASSERT(x)
#endif

Mutex::Mutex()
{
#ifdef PTHREAD
	int status;
	status = pthread_mutex_init(&m_mutex, NULL);
	if( status != 0 )
	{
		throw ThreadException(strerror(status), 
			ThreadException::MUTEX_INIT);
	}
#elif defined(WIN32)
	m_mutex = -1;
#endif
}

Mutex::~Mutex()
{
#ifdef PTHREAD
	pthread_mutex_destroy(&m_mutex);
#endif
}

void Mutex::Lock()
{
#ifdef PTHREAD
	int status;
	status = pthread_mutex_lock(&m_mutex);
	if(status != 0)
	{
		throw ThreadException(strerror(status), 
			ThreadException::MUTEX_LOCK);
	}
#elif defined(WIN32)
	LONG result;
	while(true)
	{	
		result = InterlockedIncrement(&m_mutex) ;
		if( result == 0 )
		{
			break;
		}
		else if( result > 0 )
		{
			--m_mutex;
		}
		else
		{
			ASSERT(0); // It shouldn't be less than zero ever
		}
		Sleep(1);
	}
#endif
}

bool Mutex::TryLock()
{
#ifdef PTHREAD
	int status = pthread_mutex_trylock(&m_mutex);
	if(status == EBUSY)
	{
		return false;
	}
	else if(status != 0)
	{
		throw ThreadException(strerror(status),
							  ThreadException::MUTEX_TRYLOCK);
	}
	return true;
#elif defined(WIN32)
	LONG result;
	result = InterlockedIncrement(&m_mutex) ;
	if( result == 0 )
	{
		return true;
	}
	else if( result > 0 )
	{
		--m_mutex;
	}
	return false;
#endif
}

void Mutex::Unlock()
{
#ifdef PTHREAD
	int status;
	status = pthread_mutex_unlock(&m_mutex);
	if( status != 0 )
	{
		throw ThreadException(strerror(status), 
			ThreadException::MUTEX_UNLOCK);
	
	}
#elif defined(WIN32)
	ASSERT(m_mutex >= 0);
	--m_mutex;
#endif
}

Conditional::Conditional()
{
#ifdef PTHREAD
	int status;
	status = pthread_cond_init(&m_cond, NULL);

	if(status != 0)
	{
		throw ThreadException(strerror(status), ThreadException::COND_CREATE);
	}
#endif
}

Conditional::~Conditional()
{
#ifdef PTHREAD
	int status;
	status = pthread_cond_destroy(&m_cond);
	if(status != 0)
	{
		throw ThreadException(strerror(status), ThreadException::COND_DESTROY);
							  
	}
#endif
}

void Conditional::Wait(Mutex *mutex)
{
#ifdef PTHREAD
	int status;

	status = pthread_cond_wait(&m_cond, &(mutex->m_mutex) );
	if(status != 0)
	{
		throw ThreadException(strerror(status), ThreadException::COND_WAIT);
	}
#elif defined(WIN32)
	// Create the mutex that will be unlocked when we are signaled
	Mutex notify; 
	// Lock it right away, so that the next call will block until someone
	// else unlocks it.
	notify.Lock();

	// Unlock the resource
	mutex->Unlock();

	// Add the notification mutex to the queue.
	m_wait_list_mutex.Lock();
	m_wait_list.push(&notify);
	m_wait_list_mutex.Unlock();

	// Wait until we are signaled
	notify.Lock();

	// Relock the resource
	mutex->Lock();
#endif
}

void Conditional::Signal()
{
#ifdef PTHREAD
	int status;
	status = pthread_cond_signal(&m_cond);
	if( status != 0 )
	{
		throw ThreadException(strerror(status), ThreadException::COND_SIGNAL);
	}
#elif defined(WIN32)

	// Remove the front of the queue
	m_wait_list_mutex.Lock();
	if( !m_wait_list.empty())
	{
		Mutex *notify = m_wait_list.front();
		m_wait_list.pop();
		
		// Notify the thread waiting
		notify->Unlock();
	}
	m_wait_list_mutex.Unlock();
#endif
}

void Conditional::Broadcast()
{
#ifdef PTHREAD
	int status;
	status = pthread_cond_broadcast(&m_cond);
	if(status != 0)
	{
		throw ThreadException(strerror(status), ThreadException::COND_BCAST);
	}
#elif defined(WIN32)
	Mutex *notify;
	m_wait_list_mutex.Lock();
	// For each waiting thread 
	while(!m_wait_list.empty())
	{
		notify = m_wait_list.front();
		m_wait_list.pop();
		notify->Unlock();
	}
	m_wait_list_mutex.Unlock();

#endif
}

Messenger::Messenger()
{
}

Messenger::~Messenger()
{
}

void Messenger::SndMessage(Message *msg)
{
	m_message_queue_mutex.Lock();
	m_message_queue.push(msg);
	m_message_queue_mutex.Unlock();

	m_message_queue_cond.Signal();
}

Message *Messenger::GetMessage(bool block)
{
	Message *msg;
	m_message_queue_mutex.Lock();
	if(m_message_queue.empty())
	{
		if( block )
		{
			while(m_message_queue.empty())
			{
				m_message_queue_cond.Wait(&m_message_queue_mutex);
			}
		}
		else
		{
			m_message_queue_mutex.Unlock();
			return NULL;
		}
	}

	msg = m_message_queue.front();
	m_message_queue.pop();
	m_message_queue_mutex.Unlock();
	return msg;
}


#ifdef PTHREAD
map<Thread *, Mutex *> Thread::ms_thread_lock;
#endif

Thread::Thread()
{
	m_disowned = false;
	m_reaped = true;
	m_parent_thread_id = GetThreadID();
	m_child_thread_id = 0;
	m_running = false;
	m_killed = true;
#ifdef PTHREAD
	ms_thread_lock[this] = 0;
#endif
}

Thread::~Thread()
{
#ifdef PTHREAD
	pthread_t tid = pthread_self();
	if(tid == m_child_thread_id)
	{
#ifdef _DEBUG
		printf("Thread %d exiting.\n", m_child_thread_id);
#endif
		pthread_exit(0);
	}
	else
	{
		if(!m_disowned && !m_reaped && !m_killed)
		{
			Reap();
		}
		else 
		{
			if(Running())
			{
				Kill();
			}
		}
	}
#elif defined(WIN32)
	DWORD tid = GetCurrentThreadId();
	if(tid == m_child_thread_id)
	{
#ifdef _DEBUG
		printf("Thread %d exiting.\n", m_child_thread_id);
#endif
		ExitThread(0);
	}
	else
	{
		if(!m_disowned && Running())
		{
			Reap();
		}
		else if(Running())
		{
			Kill();
		}
		CloseHandle(m_child_thread_handle);
	}
#endif
}

void Thread::Run()
{

#ifdef PTHREAD
	pthread_attr_t attr;
	int status;

	status = pthread_attr_init (&attr);
	if( status != 0 )
	{
		throw ThreadException(strerror(status),
				  ThreadException::THREAD_ATTRINIT);
	}

	status = pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	if( status == EPERM )
	{
#ifdef DEBUG
		printf("System level thread failed, trying process level...\n");
#endif
		status = pthread_attr_setscope(&attr, PTHREAD_SCOPE_PROCESS);
	}

	if( status != 0 )
	{
	
		throw ThreadException(strerror(status),
				  ThreadException::THREAD_SETSCOPE);
	}

	m_running = true;	
	m_reaped = false;
	m_killed = false;

	if( ms_thread_lock[this] == 0 )
	{
		ms_thread_lock[this] = new Mutex();
	}
	else
	{
		ms_thread_lock[this]->Unlock();
	}

	status = pthread_create(&m_child_thread_id, &attr, thread_func, this);

	if( status != 0 )
	{
		throw ThreadException(strerror(status), 
							  ThreadException::THREAD_CREATE);		
	}

	pthread_attr_destroy(&attr);
#elif defined(WIN32)
	m_child_thread_handle = CreateThread(NULL, 0, 
		(unsigned long (__stdcall *)(void *))thread_func, this, 0, &m_child_thread_id);

	if(m_child_thread_handle == NULL)
	{

		throw ThreadException("CreateThread() failed", 
							  ThreadException::THREAD_CREATE);
	}
#endif
	
#ifdef _DEBUG
	printf("Started thread %d.\n", m_child_thread_id);
#endif
}

void Thread::Reap()
{
	// If we disowned the thread, there is nothing to reap
	if(m_disowned || m_reaped) 
	{
		return;	
	}
#ifdef PTHREAD
	int status;
	void *foo;
	ASSERT(pthread_self() != m_child_thread_id);

	status = pthread_join(m_child_thread_id, &foo);
	if( status != 0 )
	{
		throw ThreadException(strerror(status), ThreadException::THREAD_JOIN);
	}
#elif defined(WIN32)
	DWORD result;

	ASSERT(GetCurrentThreadId() != m_child_thread_id);

	result = WaitForSingleObject(m_child_thread_handle, INFINITE);

	ASSERT(result != WAIT_TIMEOUT);
#endif
	m_reaped = true;
#ifdef _DEBUG
	printf("Reaped() thread %d\n", m_child_thread_id);
#endif
}

void Thread::Kill()
{
	if(!Running())
	{
		return;
	}
#ifdef PTHREAD
	int status;

	status = pthread_cancel(m_child_thread_id);
	if( status != 0 )
	{
		throw ThreadException(strerror(status), 
							  ThreadException::THREAD_KILL);

	}
#elif defined(WIN32)
	BOOL result;
	
	result = TerminateThread(m_child_thread_handle, 0);
	if( result == 0 )
	{
		throw ThreadException("TerminateThread() failed.\n", 
			ThreadException::THREAD_KILL);
	}
#endif

#ifdef _DEBUG
	printf("Killed() thread %d\n", m_child_thread_id);
#endif
	m_running = false;
	m_killed = true;
#ifdef PTHREAD
	// Wait until the child process gives up its lifetime lock.
	ms_thread_lock[this]->Lock();
#endif
}

void Thread::Disown()
{
#ifdef PTHREAD
	int status;

	status = pthread_detach(m_child_thread_id);
	if( status != 0 )
	{
		throw ThreadException(strerror(status), 
							  ThreadException::THREAD_DISOWN);
	}
#endif
	m_disowned = true;
#ifdef _DEBUG
	printf("Disowned thread %d\n", m_child_thread_id);
#endif
}

void *Thread::thread_func(void *arg)
{
	void *return_val;
#ifdef PTHREAD
	ms_thread_lock[(Thread *)arg]->Lock();
	pthread_cleanup_push(Thread::cleanup_handler, arg);
#endif
	return_val = ((Thread *)arg)->ThreadMain();
	((Thread *)arg)->m_running = false;
#ifdef PTHREAD
	pthread_cleanup_pop(1);
#endif
	return return_val;
}

void Thread::cleanup_handler(void *arg)
{
#ifdef PTHREAD
	ms_thread_lock[(Thread *)arg]->Unlock();
#endif
}

void *Thread::ThreadMain()
{
	return 0;
}

bool Thread::Running()
{
	return m_running;
}

void Thread::Sleep(int sec, int usec)
{
#ifdef WIN32
	::Sleep(sec * 1000 + usec / 1000 );
#else
	struct timeval tv;
	tv.tv_sec = sec;
	tv.tv_usec = usec;
	select( 0, 0, 0, 0, &tv);
#endif
}

ThreadException::ThreadException(char *why, THREAD_ERROR_CODE code)
{
	m_why = new char[ strlen(why) + 1 ];
	strcpy(m_why, why);
	m_code = code;
}

ThreadException::~ThreadException()
{
	delete [] m_why;
}

void ThreadException::Handle()
{
	fprintf(stderr, "ERROR: %s\n", m_why);
}

